from django.urls import reverse

import pytest

from mozilla_django_oidc_db.models import (
    OpenIDConnectConfig,
    UserInformationClaimsSources,
)

from .utils import keycloak_login

KEYCLOAK_BASE_URL = "http://localhost:8080/realms/test/"


@pytest.mark.vcr
def test_client_id_secret_full_flow(
    keycloak_config, mock_state_and_nonce, client, django_user_model, vcr
):
    login_url = reverse("login")
    django_login_response = client.get(login_url)
    assert django_login_response.status_code == 302

    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"])

    # complete the login flow on our end
    callback_response = client.get(redirect_uri)

    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    # a user was created
    assert django_user_model.objects.count() == 1

    # check that the token request was performed as expected
    token_request = next(
        req
        for req in vcr.requests
        if req.uri == f"{KEYCLOAK_BASE_URL}protocol/openid-connect/token"
        and req.method == "POST"
    )
    assert token_request is not None
    assert b"client_id=testid" in token_request.body
    assert b"secret=7DB3KUAAizYCcmZufpHRVOcD0TOkNO3I" in token_request.body
    assert "Authorization" not in token_request.headers


@pytest.mark.vcr
def test_credentials_in_basic_auth_header(
    keycloak_config: OpenIDConnectConfig,
    mock_state_and_nonce,
    client,
    django_user_model,
    vcr,
):
    keycloak_config.oidc_token_use_basic_auth = True
    keycloak_config.save()

    django_login_response = client.get(reverse("login"))
    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"])

    # complete the login flow on our end
    callback_response = client.get(redirect_uri)
    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    # check that the token request was performed as expected
    token_request = next(
        req
        for req in vcr.requests
        if req.uri == f"{KEYCLOAK_BASE_URL}protocol/openid-connect/token"
        and req.method == "POST"
    )
    assert token_request is not None

    assert "Authorization" in token_request.headers
    bits = token_request.headers["Authorization"].split(" ")
    assert len(bits) == 2
    assert bits[0] == "Basic"

    assert b"client_id=testid" in token_request.body
    assert b"secret=" not in token_request.body


@pytest.mark.vcr
def test_return_jwt_from_userinfo_endpoint(
    keycloak_config: OpenIDConnectConfig,
    mock_state_and_nonce,
    client,
    django_user_model,
):
    # Set up client configured to return JWT from userinfo endpoint instead of plain
    # JSON. Credentials from ``docker/import`` realm export.
    keycloak_config.oidc_rp_client_id = "test-userinfo-jwt"
    keycloak_config.oidc_rp_client_secret = "ktGlGUELd1FR7dTXc84L7dJzUTjCtw9S"
    keycloak_config.userinfo_claims_source = (
        UserInformationClaimsSources.userinfo_endpoint
    )
    keycloak_config.save()

    django_login_response = client.get(reverse("login"))
    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"])

    # complete the login flow on our end
    callback_response = client.get(redirect_uri)
    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    # a user was created
    assert django_user_model.objects.count() == 1



from django.apps import AppConfig


class MozillaDjangoOidcDbConfig(AppConfig):
    name = "mozilla_django_oidc_db"
    default_auto_field = "django.db.models.AutoField"

    def ready(self) -> None:
        from . import checks  # noqa



from typing import Any

import pytest

from mozilla_django_oidc_db.config import dynamic_setting


class ConfigProvider:
    def get_settings(self, attr: str, *args: Any) -> Any: ...

    SETTING_WITHOUT_DEFAULT = dynamic_setting[str]()
    SETTING_WITH_DEFAULT = dynamic_setting(default=123)


@pytest.mark.parametrize(
    "attribute,expected",
    [
        ("SETTING_WITHOUT_DEFAULT", "<dynamic_setting SETTING_WITHOUT_DEFAULT>"),
        (
            "SETTING_WITH_DEFAULT",
            "<dynamic_setting SETTING_WITH_DEFAULT (default: 123)>",
        ),
    ],
)
def test_representation_class_attribute(attribute: str, expected: str):
    setting = getattr(ConfigProvider, attribute)

    assert repr(setting) == expected


def test_dynamic_settings_are_data_descriptors():
    instance = ConfigProvider()

    with pytest.raises(AttributeError):
        instance.SETTING_WITHOUT_DEFAULT = "something else"



# Mapping the configuration model fieldnames for endpoints to their
# corresponding names in the OIDC spec
OIDC_MAPPING = {
    "oidc_op_authorization_endpoint": "authorization_endpoint",
    "oidc_op_token_endpoint": "token_endpoint",
    "oidc_op_user_endpoint": "userinfo_endpoint",
    "oidc_op_jwks_endpoint": "jwks_uri",
    "oidc_op_logout_endpoint": "end_session_endpoint",
}

OPEN_ID_CONFIG_PATH = ".well-known/openid-configuration"



# Generated by Django 3.2.13 on 2022-04-22 08:49

from django.db import migrations, models

import mozilla_django_oidc_db.models


class Migration(migrations.Migration):
    dependencies = [
        ("mozilla_django_oidc_db", "0007_auto_20220307_1128"),
    ]

    operations = [
        migrations.AlterField(
            model_name="openidconnectconfig",
            name="claim_mapping",
            field=models.JSONField(
                default=mozilla_django_oidc_db.models.get_claim_mapping,
                help_text="Mapping from user-model fields to OIDC claims",
                verbose_name="claim mapping",
            ),
        ),
        migrations.AlterField(
            model_name="openidconnectconfig",
            name="oidc_op_jwks_endpoint",
            field=models.URLField(
                blank=True,
                help_text="URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm.",
                max_length=1000,
                verbose_name="JSON Web Key Set endpoint",
            ),
        ),
        migrations.AlterField(
            model_name="openidconnectconfig",
            name="oidc_rp_idp_sign_key",
            field=models.CharField(
                blank=True,
                help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format.",
                max_length=1000,
                verbose_name="Sign key",
            ),
        ),
    ]



# Generated by Django 2.2.24 on 2021-11-23 14:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mozilla_django_oidc_db", "0004_auto_20210812_1044"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="sync_groups_glob_pattern",
            field=models.CharField(
                default="*",
                help_text="The glob pattern that groups must match to be synchronized to the local database.",
                max_length=255,
                verbose_name="groups glob pattern",
            ),
        ),
    ]



# Generated by Django 2.2.24 on 2021-07-16 12:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("mozilla_django_oidc_db", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="oidc_op_discovery_endpoint",
            field=models.URLField(
                blank=True,
                help_text="URL of your OpenID Connect provider discovery endpoint. If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                max_length=1000,
                verbose_name="Discovery endpoint",
            ),
        ),
    ]



import logging
from collections.abc import Collection
from copy import deepcopy

import requests
from glom import Path, PathAccessError, assign, glom
from requests.utils import _parse_content_type_header  # type: ignore

from .models import OpenIDConnectConfigBase
from .typing import ClaimPath, JSONObject, JSONValue

logger = logging.getLogger(__name__)


def obfuscate_claim_value(value: JSONValue) -> JSONValue:
    """
    Obfuscates the value of a claim, so it can be logged safely

    If a dict of claims is supplied, all of the values will be obfuscated
    """
    if isinstance(value, dict):
        for k, v in value.items():
            value[k] = obfuscate_claim_value(v)
        return value
    else:
        value = str(value)
        threshold = int(len(value) * 0.75)
        return "".join([x if i > threshold else "*" for i, x in enumerate(value)])


def obfuscate_claims(
    claims: JSONObject, claims_to_obfuscate: Collection[ClaimPath]
) -> JSONObject:
    """
    Obfuscates the specified claims in the provided claims object.
    """
    copied_claims = deepcopy(claims)
    for claim_bits in claims_to_obfuscate:
        claim_path = Path(*claim_bits)
        try:
            claim_value = glom(copied_claims, claim_path)
        except PathAccessError:
            continue
        assign(copied_claims, claim_path, obfuscate_claim_value(claim_value))
    return copied_claims


def extract_content_type(ct_header: str) -> str:
    """
    Get the content type + parameters from content type header.

    This is internal API since we use a requests internal utility, which may be
    removed/modified at any time. However, this is a deliberate choices since I trust
    requests to have a correct implementation more than coming up with one myself.
    """
    content_type, _ = _parse_content_type_header(ct_header)
    # discard the params, we only want the content type itself
    return content_type


def do_op_logout(config: OpenIDConnectConfigBase, id_token: str) -> None:
    """
    Perform the logout with the OpenID Provider.

    Standard: https://openid.net/specs/openid-connect-rpinitiated-1_0.html#RPLogout

    .. warning:: Preferably, you should send the user to the configured logout endpoint
       so they can confirm the logout and any session cookies are cleared. If that is not
       possible, you can call this helper for server-to-server logout, but there are no
       guarantees this works for every possible OpenID Provider implementation. It has
       been tested with Keycloak, but the standard says nothing about server-to-server
       calls to log out a user.
    """
    logout_endpoint = config.oidc_op_logout_endpoint
    if not logout_endpoint:
        return

    response = requests.post(logout_endpoint, data={"id_token_hint": id_token})
    if not response.ok:
        logger.warning(
            "Failed to log out the user at the OpenID Provider. Status code: %s",
            response.status_code,
            extra={
                "response": response,
                "status_code": response.status_code,
            },
        )



from django.test import Client
from django.urls import reverse

import pytest
from requests import Session

from mozilla_django_oidc_db.models import OpenIDConnectConfig
from mozilla_django_oidc_db.utils import do_op_logout

from .utils import keycloak_login


@pytest.fixture
def kc_session(
    settings,
    keycloak_config,
    mock_state_and_nonce,
    client,
    django_user_model,
    vcr,
):
    settings.OIDC_STORE_ID_TOKEN = True
    session = Session()

    login_url = reverse("login")
    django_login_response = client.get(login_url)
    assert django_login_response.status_code == 302

    # simulate login to Keycloak
    redirect_uri = keycloak_login(django_login_response["Location"], session=session)

    # complete the login flow on our end
    callback_response = client.get(redirect_uri)

    assert callback_response.status_code == 302
    assert callback_response["Location"] == "/admin/"

    # a user was created
    assert django_user_model.objects.count() == 1

    # assert that we are logged in to keycloak
    django_login_response2 = client.get(login_url)

    kc_response = session.get(django_login_response2["Location"], allow_redirects=False)
    assert kc_response.status_code == 302
    assert kc_response.headers["Location"].startswith("http://testserver")

    yield (client, session)

    session.close()


@pytest.mark.vcr
@pytest.mark.oidcconfig(oidc_op_logout_endpoint="")
def test_logout_without_endpoint_configured(
    keycloak_config: OpenIDConnectConfig,
    kc_session: tuple[Client, Session],
):
    client, session = kc_session

    do_op_logout(keycloak_config, id_token=client.session["oidc_id_token"])

    # check that we are still authenticated in keycloak
    login_url = reverse("login")
    django_login_response = client.get(login_url)
    kc_response = session.get(django_login_response["Location"], allow_redirects=False)

    assert kc_response.status_code == 302
    assert kc_response.headers["Location"].startswith("http://testserver")


@pytest.mark.vcr
def test_logout_with_logout_endpoint_configured(
    keycloak_config: OpenIDConnectConfig,
    kc_session: tuple[Client, Session],
):
    assert keycloak_config.oidc_op_logout_endpoint
    client, session = kc_session

    do_op_logout(keycloak_config, id_token=client.session["oidc_id_token"])

    # check that we are still authenticated in keycloak
    login_url = reverse("login")
    django_login_response = client.get(login_url)
    kc_response = session.get(django_login_response["Location"], allow_redirects=False)

    assert kc_response.status_code == 200, "Did not end up on Keycloak's login page"
    assert kc_response.headers["Content-Type"].startswith("text/html")



"""
Test the OIDC Authenticaton Request flow with our custom views.
"""

from urllib.parse import parse_qs, urlsplit

from django.urls import reverse

import pytest

from mozilla_django_oidc_db.exceptions import OIDCProviderOutage


@pytest.mark.oidcconfig(
    oidc_op_authorization_endpoint="http://localhost:8080/openid-connect/auth"
)
def test_default_config_flow(dummy_config, client):
    start_url = reverse("oidc_authentication_init")
    assert start_url == "/oidc/authenticate/"

    response = client.get(start_url, {"next": "/admin/"})

    # check that the view redirects to the identity provider
    assert response.status_code == 302
    parsed_url = urlsplit(response.url)
    assert parsed_url.scheme == "http"
    assert parsed_url.netloc == "localhost:8080"
    assert parsed_url.path == "/openid-connect/auth"

    # introspect state
    state_key = parse_qs(parsed_url.query)["state"][0]
    state = client.session["oidc_states"][state_key]
    assert state["config_class"] == "mozilla_django_oidc_db.OpenIDConnectConfig"
    # upstream library
    assert client.session["oidc_login_next"] == "/admin/"
    # our own addition
    assert client.session["oidc-db_redirect_next"] == "/admin/"


@pytest.mark.oidcconfig(oidc_keycloak_idp_hint="keycloak-idp2")
def test_keycloak_idp_hint_via_config(dummy_config, settings, client):
    settings.OIDC_KEYCLOAK_IDP_HINT = "keycloak-idp1"
    start_url = reverse("oidc_authentication_init")

    response = client.get(start_url)

    assert response.status_code == 302
    parsed_url = urlsplit(response.url)

    query = parse_qs(parsed_url.query)
    assert query["kc_idp_hint"] == ["keycloak-idp2"]


def test_check_idp_availability_not_available(
    dummy_config, settings, client, requests_mock
):
    settings.OIDCDB_CHECK_IDP_AVAILABILITY = True
    requests_mock.get("https://mock-oidc-provider:9999/oidc/auth", status_code=503)
    start_url = reverse("oidc_authentication_init")

    with pytest.raises(OIDCProviderOutage):
        client.get(start_url)


def test_check_idp_availability_available(
    dummy_config, settings, client, requests_mock
):
    settings.OIDCDB_CHECK_IDP_AVAILABILITY = True
    requests_mock.get("https://mock-oidc-provider:9999/oidc/auth", status_code=400)
    start_url = reverse("oidc_authentication_init")

    response = client.get(start_url)

    assert response.status_code == 302



"""
Test the init flow through a custom view and config.
"""

from urllib.parse import parse_qs, urlsplit

from django.core.exceptions import DisallowedRedirect

import pytest

from mozilla_django_oidc_db.exceptions import OIDCProviderOutage
from mozilla_django_oidc_db.views import OIDCInit

from .custom_config import CustomConfig, oidc_init

pytestmark = [pytest.mark.django_db]


def test_redirects_to_oidc_provider(auth_request):
    response = oidc_init(auth_request, return_url="/fixed-next")

    assert response.status_code == 302
    parsed_url = urlsplit(response.url)
    assert parsed_url.scheme == "https"
    assert parsed_url.netloc == "example.com"
    assert parsed_url.path == "/oidc/auth"

    # introspect state
    state_key = parse_qs(parsed_url.query)["state"][0]
    state = auth_request.session["oidc_states"][state_key]
    assert state["config_class"] == "mozilla_django_oidc_db.CustomConfig"
    # upstream library
    assert auth_request.session["oidc_login_next"] == "/fixed-next"
    # our own addition
    assert auth_request.session["oidc-db_redirect_next"] == "/fixed-next"


def test_suspicious_return_url(auth_request):
    with pytest.raises(DisallowedRedirect):
        oidc_init(auth_request, return_url="http://evil.com/steal-my-data")


@pytest.mark.parametrize(
    "get_kwargs",
    (
        {},
        {"return_url": ""},
        {"return_url": None},
    ),
)
def test_forgotten_return_url(auth_request, get_kwargs):
    with pytest.raises(ValueError):
        oidc_init(auth_request, **get_kwargs)


class IDPCheckInitView(OIDCInit):
    def check_idp_availability(self) -> None:
        raise OIDCProviderOutage("The internet is bwoken.")


oidc_init_with_idp_check = IDPCheckInitView.as_view(
    config_class=CustomConfig, allow_next_from_query=True
)


def test_idp_check_mechanism(auth_request, settings):
    settings.OIDCDB_CHECK_IDP_AVAILABILITY = True

    with pytest.raises(OIDCProviderOutage):
        oidc_init_with_idp_check(auth_request)
